if(typeof VTL === "undefined") VTL = {};
VTL.OmniViewer = class {
	canvas;

	width;
	height;

	// reuse items
	mouse; // vec2;
	raycaster;
	objectPoint; // vec3
	
	camera;
	sphere;
	scene;
	renderer;
	materials = {};

	control;
	
	onRequestAnimation;

	constructor(canvas){
		this.width = canvas.width;
		this.height = canvas.height;

		this.canvas = canvas;

		this.renderer = new THREE.WebGLRenderer({canvas});
		this.renderer.setClearColor({color: 0x00ff00});
		this.renderer.setSize(this.width, this.height);

		this.scene = new THREE.Scene();
		this.scene.background = new THREE.Color( 0xffffff );

		this.sphere = new THREE.SphereGeometry(10, 600, 600);
		this.sphere.scale(-1,1,1);

		this.mesh = new THREE.Mesh(this.sphere, null);
		this.mesh.matrixAutoUpdate = false; // object is static
		this.scene.add(this.mesh);

		const fov = 90;
		const aspect = this.width / this.height;
		const near = 0.1;
		const far = 1000;
		this.camera = new THREE.PerspectiveCamera(fov, aspect, near, far);
		this.camera.position.set(0, 0, 0);
		this.scene.add(this.camera);
		
		// activate mouse interaction
		this.control = new VTL.CameraControl(this.camera, this.canvas);
		this.control.resetPosition();

		this.mouse = new THREE.Vector2();
		this.objectPoint = new THREE.Vector3();
		this.raycaster = new THREE.Raycaster();

		this.onRequestAnimation = null;
	}
}


VTL.OmniViewer.prototype.render = function(){
	let self = this;

	if(this.mesh.material != null){
		this.renderer.render(this.scene, this.camera);
		if(this.onRequestAnimation){
			this.onRequestAnimation();
		}
	}
	requestAnimationFrame(function(){self.render()});
}

VTL.OmniViewer.prototype.updateSize = function(width, height)
{
	this.renderer.setSize(width, height);

	this.camera.aspect = width/height;
	this.camera.updateProjectionMatrix();

	//this.canvas.width = width;
	//this.canvas.height = height;

	this.width = width;
	this.height = height;
}

// call only il-material
VTL.OmniViewer.prototype.updateRefRatio = function(material, value){
	material.uniforms.uRefRatio.value = value;
}
// call only scaling changeable images
VTL.OmniViewer.prototype.updateScaling = function(material, min, max){
	material.uniforms.uMin.value = min;
	material.uniforms.uMax.value = max;
}

VTL.OmniViewer.prototype.setMaterial = function(mat){
	this.mesh.material = mat;
	if(mat != null){
		this.mesh.material.needsUpdate = true;
	}
}

// for normal
VTL.OmniViewer.createTexture = function(imgSrc){
	return new THREE.TextureLoader().load(imgSrc);
}

VTL.OmniViewer.createDataTexture = function(float32ary, width, height){
	let texture = new THREE.DataTexture(float32ary, width, height, THREE.RedFormat, THREE.FloatType);
	texture.flipY = true; // 
	return texture;
}

VTL.OmniViewer.createMaterial = function(texture){
	return new THREE.MeshBasicMaterial({map:texture});
}

VTL.OmniViewer.createLMMaterial = function(lmTexture, min, max){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTex: {type: "t", value: lmTexture },
			uMin: {type:"f", value: min },
			uMax: {type:"f", value: max },
		},
		vertexShader: 
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTex;
uniform float uMin;
uniform float uMax;

float log10(float x){
	return 0.43429448190325176 * log(x);
}

void main() {
	float value = float(texture2D(uTex, vUv));
	value = value == 0.0 ? 0.000001 : value;
	value = max(uMin, min(uMax, value));
	float ratio = (log10(value)-log10(uMin))/(log10(uMax) - log10(uMin));
	float t = 1.0 - (ratio * 2.0);

	gl_FragColor = vec4(
		clamp(1.5 - abs(2.0 * t + 1.0), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t - 1.0), 0.15, 0.85),
		1.0);
}
`
	});
}

VTL.OmniViewer.createILMaterial = function(lmTexture, min, max, refRatio){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTex: {type: "t", value: lmTexture },
			uMin: {type:"f", value: min},
			uMax: {type:"f", value: max},
			uRefRatio: {type:"f", value: refRatio},
		},
		vertexShader: 
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTex;
uniform float uMin;
uniform float uMax;
uniform float uRefRatio;

const float PI = 3.1415926535;

float log10(float x){
	return 0.43429448190325176 * log(x);
}

void main() {
	float lm = float(texture2D(uTex, vUv));
	float il = lm * PI / uRefRatio;
	il = il == 0.0 ? 0.000001 : il;
	il = max(uMin, min(uMax, il));
	float ratio = (log10(il)-log10(uMin))/(log10(uMax) - log10(uMin));
	float t = 1.0-(ratio * 2.0);

	gl_FragColor = vec4(
		clamp(1.5 - abs(2.0 * t + 1.0), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t - 1.0), 0.15, 0.85),
		1.0);
}
`
	});
}

VTL.OmniViewer.createBRMaterial = function(brTexture, min, max){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTex: {type: "t", value: brTexture },
			uMin: {type:"f", value: min},
			uMax: {type:"f", value: max},
		},
		vertexShader:
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTex;
uniform float uMin;
uniform float uMax;

void main() {
	float value = float(texture2D(uTex, vUv));
	value = max(uMin, min(uMax, value));
	float ratio = (value-uMin)/(uMax - uMin);
	float t = 1.0-(ratio * 2.0);
	gl_FragColor = vec4(
		clamp(1.5 - abs(2.0 * t + 1.0), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t - 1.0), 0.15, 0.85),
		1.0);
}
`
	});
}

VTL.OmniViewer.createEBRMaterial = function(brTexture){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTex: {type: "t", value: brTexture }
		},
		vertexShader:
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTex;

void main() {
	float value = float(texture2D(uTex, vUv));
	vec4 color;
	if(value >= 9.0){
		color = vec4(1.0, 0.0, 0.0, 1.0);
	} 
	else if( value >= 7.5){
		color = vec4(1.0, 1.0, 0.0, 1.0);
	}
	else if( value >= 6.5){
		color = vec4(0.0, 1.0, 0.0, 1.0);
	}
	else if( value >= 4.0){
		color = vec4(0.0, 0.0, 1.0, 1.0);
	}
	else{
		color = vec4(0.0, 0.0, 0.0, 1.0);
	}
		
	gl_FragColor = color;
}
`
	});
}

VTL.OmniViewer.createUGRMaterial = function(raTex, ugrTex, min, max){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTexRa: {type: "t", value: raTex },
			uTexUgr: {type: "t", value: ugrTex },
			uMin: {type:"f", value: min},
			uMax: {type:"f", value: max},
		},
		transparent: true,
		vertexShader:
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTexRa;
uniform sampler2D uTexUgr;
uniform float uMin;
uniform float uMax;

void main() {
	vec4 raColor = vec4(texture2D(uTexRa, vUv).rgb, 0.8);
	float value = float(texture2D(uTexUgr, vUv));
	value = max(uMin, min(uMax, value));
	float ratio = (value-uMin)/(uMax - uMin);
	float t = 1.0-(ratio * 2.0);

	vec4 ugrColor = vec4(
		clamp(1.5 - abs(2.0 * t + 1.0), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t), 0.15, 0.85),
		clamp(1.5 - abs(2.0 * t - 1.0), 0.15, 0.85),
		1.0);

	gl_FragColor = value > uMin ? ugrColor : raColor;
}
`
	});
}

VTL.OmniViewer.createLabMaterial = function(raTex, lTex, aTex, bTex, chroma, brightness){
	return new THREE.ShaderMaterial({
		uniforms:{
			uTexRa: {type: "t", value: raTex },
			uTexL: {type: "t", value: lTex },
			uTexA: {type: "t", value: aTex },
			uTexB: {type: "t", value: bTex },
			uChromaLim: {type: "f", value: chroma },
			uBrightnessLim: {type: "f", value: brightness },
		},
		transparent: true,
		vertexShader:
`
varying vec2 vUv;
void main() {
	vUv = uv;
	vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
	gl_Position = projectionMatrix * mvPosition;
}
`,
		fragmentShader:
`
varying vec2 vUv;
uniform sampler2D uTexRa;
uniform sampler2D uTexL;
uniform sampler2D uTexA;
uniform sampler2D uTexB;
uniform float uChromaLim;
uniform float uBrightnessLim;

const float PI = 3.1415926535;

float atan2(in float y, in float x)
{
    return x == 0.0 ? sign(y)*PI/2.0 : atan(y, x);
}

void main() {
	vec4 raColor = vec4(texture2D(uTexRa, vUv).rgb, 0.8);
	float l = float(texture2D(uTexL, vUv));
	float a = float(texture2D(uTexA, vUv));
	float b = float(texture2D(uTexB, vUv));
	float distance = sqrt((a * a) + (b * b));

	vec3 color = raColor.rgb;

	float deg = 180.0 * atan2(b, a) / PI;
	deg = deg < 0.0 ? 360.0 + deg: deg;
	if(deg < 36.0) color = vec3(1.0000, 0.1607, 0.2901);
	else if(deg < 72.0)  color = vec3(1.0000, 0.5411, 0.0000);
	else if(deg < 108.0) color = vec3(1.0000, 0.8352, 0.0000);
	else if(deg < 144.0) color = vec3(0.8156, 1.0000, 0.0000);
	else if(deg < 180.0) color = vec3(0.0000, 1.0000, 0.7333);
	else if(deg < 216.0) color = vec3(0.0000, 1.0000, 0.9607);
	else if(deg < 252.0) color = vec3(0.0000, 0.9803, 1.0000);
	else if(deg < 288.0) color = vec3(0.0000, 0.5843, 1.0000);
	else if(deg < 324.0) color = vec3(0.8352, 0.4784, 1.0000);
	else if(deg < 360.0) color = vec3(1.0000, 0.3254, 0.6666);
	
	gl_FragColor = uChromaLim <= distance && uBrightnessLim <= l ? vec4(color, 1.0): raColor;
}
`
	});
}

VTL.OmniViewer.prototype.locationToSphereInfo = function(offsetX, offsetY){
	this.mouse.set((offsetX / this.width) * 2 - 1, -(offsetY / this.height) * 2 + 1);
	this.raycaster.setFromCamera( this.mouse, this.camera );
	let obj = this.raycaster.intersectObjects(this.scene.children)[0];
	return {point:obj.point, uv:obj.uv};
}


VTL.OmniViewer.prototype.worldPointToClientLocation = function(x, y, z){
	this.objectPoint.set(x, y, z);

	const projection = this.objectPoint.project(this.camera);
	const sx = (this.width * 0.5) * (+projection.x + 1.0);
	const sy = (this.height * 0.5) * (-projection.y + 1.0);
	
	return {x: sx, y:sy};
}
VTL.OmniViewer.prototype.resetPosition = function(){
	this.control.resetPosition();
}
